import glob
import os
import re
from abc import ABC, abstractmethod
from io import BytesIO

import boto3

from .misc import LogFolderType


class FileHandler(ABC):
    """An abstract base class that encapsulates the interaction with a model storage.

    Currently implemented trough `FSFileHandler` and `S3FileHandler`.

    Methods are exposed to detect what kind of model folder is present, to list files
    and to read files into a buffer.
    """
    type: LogFolderType = LogFolderType.UNKNOWN_FOLDER
    training_simtrace_path: str = None
    training_simtrace_split: str = None
    training_robomaker_log_path: str = None
    evaluation_simtrace_path: str = None
    evaluation_robomaker_log_path: str = None
    evaluation_robomaker_split: str = None
    evaluation_simtrace_split: str = None
    leaderboard_robomaker_log_path: str = None
    leaderboard_robomaker_log_split: str = None
    model_metadata_path: str = None
    hyperparameters_path: str = None

    @abstractmethod
    def list_files(self, filterexp: str = None, check_exist: bool = False) -> list:
        pass

    @abstractmethod
    def get_file(self, key: str) -> bytes:
        pass

    @abstractmethod
    def determine_root_folder_type(self) -> LogFolderType:
        pass


class FSFileHandler(FileHandler):
    """An implementation of the FileHandler interface that interacts with the local filesystem.

    Methods are exposed to detect what kind of model folder is present, to list files
    and to read files into a buffer.
    """
    def __init__(self, model_folder: str, simtrace_path=None, robomaker_log_path=None):
        self.model_folder = model_folder
        self.training_simtrace_path = simtrace_path
        self.training_robomaker_log_path = robomaker_log_path

    def list_files(self, filterexp: str = None, check_exist: bool = False) -> list:
        """Lists a set of files.

        If `filterexp` is provided then a search is performed using provided pattern.
        Otherwise all files in the `model_folder` is returned.

        Args:
            filterexp:
                Filter expression. Typically the values for `training_simtrace_path`,
                `evaluation_simtrace_path` etc. are provided.
            check_exist:
                If `True` then raise Exception if no files are found.
                Otherwise return empty list.

        Returns:
            A list of files.
        """
        if check_exist and (filterexp is None and self.model_folder is None):
            raise Exception("File path is not defined.")

        if filterexp is None:
            return_files = glob.glob(self.model_folder)
        else:
            return_files = glob.glob(filterexp)

        if len(return_files) > 0:
            return return_files
        else:
            if check_exist:
                raise Exception("No files found in {} or {}".format(self.model_folder, filterexp))
            else:
                return []

    def get_file(self, key: str) -> bytes:
        """Downloads a given gile as byte array.

        Args:
            key:
                Path to a file on the filesystem.

        Returns:
            A bytes object containing the file.
        """

        bytes_io: BytesIO = None
        with open(key, 'rb') as fh:
            bytes_io = BytesIO(fh.read())
        return bytes_io.getvalue()

    def determine_root_folder_type(self) -> LogFolderType:

        if os.path.isdir(os.path.join(self.model_folder, "sim-trace")):
            self.type = LogFolderType.CONSOLE_MODEL_WITH_LOGS
            if self.training_simtrace_path is None:
                self.training_simtrace_path = os.path.join(
                    self.model_folder, "sim-trace", "training", "training-simtrace",
                    "*-iteration.csv")
                self.training_simtrace_split = r'(.*)/training/training-simtrace/(.*)-iteration.csv'
                self.evaluation_simtrace_path = os.path.join(
                    self.model_folder, "sim-trace", "evaluation", "*", "evaluation-simtrace",
                    "0-iteration.csv")
                self.evaluation_simtrace_split = \
                    r'.*/evaluation/([0-9]{14})-.*/evaluation-simtrace/(.*)-iteration\.csv'
            if self.training_robomaker_log_path is None:
                self.training_robomaker_log_path = glob.glob(os.path.join(
                    self.model_folder, "**", "training", "*-robomaker.log"))[0]

            if self.evaluation_robomaker_log_path is None:
                self.evaluation_robomaker_log_path = os.path.join(
                    self.model_folder, "**", "evaluation", "*-robomaker.log")

            if self.evaluation_robomaker_split is None:
                self.evaluation_robomaker_split = \
                    r'.*/evaluation/evaluation-([0-9]{14})-(.*)-robomaker.log'

            if self.leaderboard_robomaker_log_path is None:
                self.leaderboard_robomaker_log_path = os.path.join(
                    self.model_folder, "**", "leaderboard", "*-robomaker.log")

            if self.leaderboard_robomaker_log_split is None:
                self.leaderboard_robomaker_log_split = \
                    r'.*/leaderboard/leaderboard-([0-9]{14})-(.*)-robomaker.log'

            if self.model_metadata_path is None:
                self.model_metadata_path = os.path.join(
                    self.model_folder, "model", "model_metadata.json")

            if self.hyperparameters_path is None:
                self.hyperparameters_path = os.path.join(
                    self.model_folder, "ip", "hyperparameters.json")

        elif os.path.isdir(os.path.join(self.model_folder, "training-simtrace")):
            self.type = LogFolderType.DRFC_MODEL_SINGLE_WORKERS
            if self.training_simtrace_path is None:
                self.training_simtrace_path = os.path.join(
                    self.model_folder, "training-simtrace", "*-iteration.csv")
                self.training_simtrace_split = r'(.*)/training-simtrace/(.*)-iteration.csv'
                self.evaluation_simtrace_path = os.path.join(
                    self.model_folder, "evaluation-*", "evaluation-simtrace", "0-iteration.csv")
                self.evaluation_simtrace_split = \
                    r'.*/evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'

            if self.model_metadata_path is None:
                self.model_metadata_path = os.path.join(
                    self.model_folder, "model", "model_metadata.json")

            if self.hyperparameters_path is None:
                self.hyperparameters_path = os.path.join(
                    self.model_folder, "ip", "hyperparameters.json")

        elif os.path.isdir(os.path.join(self.model_folder, "0")):
            self.type = LogFolderType.DRFC_MODEL_MULTIPLE_WORKERS
            if self.training_simtrace_path is None:
                self.training_simtrace_path = os.path.join(
                    self.model_folder, "**", "training-simtrace", "*-iteration.csv")
                self.training_simtrace_split = r'.*/(.)/training-simtrace/(.*)-iteration.csv'
                self.evaluation_simtrace_path = os.path.join(
                    self.model_folder, "evaluation-*", "evaluation-simtrace", "0-iteration.csv")
                self.evaluation_simtrace_split = \
                    r'.*/evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'

            if self.model_metadata_path is None:
                self.model_metadata_path = os.path.join(
                    self.model_folder, "model", "model_metadata.json")

            if self.hyperparameters_path is None:
                self.hyperparameters_path = os.path.join(
                    self.model_folder, "ip", "hyperparameters.json")

        elif os.path.isdir(os.path.join(self.model_folder, "model")):
            self.type = LogFolderType.DRFC_MODEL_UPLOAD
            if self.training_simtrace_path is None:
                self.training_simtrace_path = None
                self.training_simtrace_split = None
                self.evaluation_simtrace_path = os.path.join(
                    self.model_folder, "evaluation-*", "evaluation-simtrace", "0-iteration.csv")
                self.evaluation_simtrace_split = \
                    r'.*/evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'

            if self.model_metadata_path is None:
                self.model_metadata_path = os.path.join(
                    self.model_folder, "model", "model_metadata.json")

            if self.hyperparameters_path is None:
                self.hyperparameters_path = os.path.join(
                    self.model_folder, "ip", "hyperparameters.json")

        return self.type


class S3FileHandler(FileHandler):
    """An implementation of the FileHandler interface that interacts with an S3 bucket.

    Methods are exposed to detect what kind of model folder is present, to list files
    and to read files into a buffer.
    """

    def __init__(self, bucket: str, prefix: str = None,
                 s3_endpoint_url: str = None, region: str = None, profile: str = None,
                 ):
        """Initializes an S3 file handler.

        The bucket can either be in real S3, or in a locally hosted S3 using minio.

        Args:
            bucket:
                Name of the bucket
            prefix:
                Prefix pointing to the root of the model folder.
            s3_endpoint_url:
                Alternate Endpoint URL, used for locally hosted S3.
            region:
                AWS region to connect to. Not applicable for locally hosted S3.
            profile:
                Name of the profile in `.aws/` which contains the credentials to use.

        """
        if profile is not None:
            session = boto3.session.Session(profile_name=profile)
            self.s3 = session.resource("s3", endpoint_url=s3_endpoint_url, region_name=region)
        else:
            self.s3 = boto3.resource("s3", endpoint_url=s3_endpoint_url, region_name=region)

        self.bucket = bucket

        if (prefix[-1:] == "/"):
            self.prefix = prefix
        else:
            self.prefix = "{}/".format(prefix)

    def list_files(self, filterexp: str = None, check_exist: bool = False) -> list:
        """Lists a set of files.

        For S3 buckets the files in `s3://{bucket}/{prefix}` is listed, and subsequently
        filtered with the expression in `filterexp`. The resulting list is returned.
        If `filterexp = None` then the entire list is returned.

        Args:
            filterexp:
                Filter expression. Typically the values for `training_simtrace_path`,
                `evaluation_simtrace_path` etc. are provided.
            check_exist:
                If `True` then raise Exception if no files are found.
                Otherwise return empty list.

        Returns:
            A list of files.
        """
        files = []

        if check_exist and filterexp is None:
            raise Exception("File path is not defined.")

        bucket_obj = self.s3.Bucket(self.bucket)
        for objects in bucket_obj.objects.filter(Prefix=self.prefix):
            files.append(objects.key)

        if filterexp is not None:
            return_files = [x for x in files if re.match(filterexp, x)]
        else:
            return_files = files

        if len(return_files) > 0:
            return return_files
        else:
            if check_exist:
                raise Exception("No files found in s3://{}/{} using filter {}"
                                .format(self.bucket, self.prefix, filterexp))
            else:
                return []

    def get_file(self, key: str) -> bytes:
        """Downloads a given gile as byte array.

        Args:
            key:
                Path to a file within a bucket.

        Returns:
            A bytes object containing the file.
        """
        bytes_io = BytesIO()
        self.s3.Object(self.bucket, key).download_fileobj(bytes_io)
        return bytes_io.getvalue()

    def determine_root_folder_type(self) -> LogFolderType:

        if len(self.list_files(filterexp=(self.prefix + r'sim-trace/(.*)'))) > 0:
            self.type = LogFolderType.CONSOLE_MODEL_WITH_LOGS
            self.training_simtrace_path = self.prefix + \
                r'sim-trace/training/training-simtrace/(.*)-iteration\.csv'
            self.training_simtrace_split = \
                r'(.*)/sim-trace/training/training-simtrace/(.*)-iteration.csv'
            self.evaluation_simtrace_path = self.prefix + \
                r'sim-trace/evaluation/(.*)/evaluation-simtrace/0-iteration\.csv'
            self.evaluation_simtrace_split = \
                r'.*sim-trace/evaluation/([0-9]{14})-.*/evaluation-simtrace/(.*)-iteration\.csv'
            self.training_robomaker_log_path = self.prefix + \
                r'logs/training/(.*)-robomaker\.log'
            self.evaluation_robomaker_log_path = self.prefix + \
                r'logs/evaluation/(.*)-robomaker\.log'
            self.evaluation_robomaker_split = \
                r'.*/evaluation/evaluation-([0-9]{14})-(.*)-robomaker.log'
            self.leaderboard_robomaker_log_path = self.prefix + \
                r'logs/leaderboard/(.*)-robomaker\.log'
            self.leaderboard_robomaker_log_split = \
                r'.*/leaderboard/leaderboard-([0-9]{14})-(.*)-robomaker.log'
            self.model_metadata_path = self.prefix + \
                r'model/model_metadata.json'
            self.hyperparameters_path = self.prefix + \
                r'ip/hyperparameters.json'

        elif len(self.list_files(filterexp=(self.prefix + r'training-simtrace/(.*)'))) > 0:
            self.type = LogFolderType.DRFC_MODEL_SINGLE_WORKERS
            self.training_simtrace_path = self.prefix + r'training-simtrace/(.*)-iteration\.csv'
            self.training_simtrace_split = r'(.*)/training-simtrace/(.*)-iteration.csv'
            self.evaluation_simtrace_path = self.prefix + \
                r'evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'
            self.evaluation_simtrace_split = \
                r'.*/evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'
            self.model_metadata_path = self.prefix + \
                r'model/model_metadata.json'
            self.hyperparameters_path = self.prefix + \
                r'ip/hyperparameters.json'

        elif len(self.list_files(filterexp=(self.prefix + r'./training-simtrace/(.*)'))) > 0:
            self.type = LogFolderType.DRFC_MODEL_MULTIPLE_WORKERS
            self.training_simtrace_path = self.prefix + \
                r'(.)/training-simtrace/(.*)-iteration\.csv'
            self.training_simtrace_split = r'.*/(.)/training-simtrace/(.*)-iteration.csv'
            self.evaluation_simtrace_path = self.prefix + \
                r'evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'
            self.evaluation_simtrace_split = \
                r'.*/evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'
            self.model_metadata_path = self.prefix + \
                r'model/model_metadata.json'
            self.hyperparameters_path = self.prefix + \
                r'ip/hyperparameters.json'

        elif len(self.list_files(filterexp=(self.prefix + r'evaluation-([0-9]{14})'))) > 0:
            self.type = LogFolderType.DRFC_MODEL_UPLOAD
            self.training_simtrace_path = None
            self.training_simtrace_split = None
            self.evaluation_simtrace_path = self.prefix + \
                r'evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'
            self.evaluation_simtrace_split = \
                r'.*/evaluation-([0-9]{14})/evaluation-simtrace/(.*)-iteration\.csv'
            self.model_metadata_path = self.prefix + \
                r'model/model_metadata.json'
            self.hyperparameters_path = self.prefix + \
                r'ip/hyperparameters.json'

        return self.type



import sys
import deepracer.boto3_enhancer


def main():
    if "install-cli" in sys.argv:
        force = True if "--force" in sys.argv else False

        deepracer.boto3_enhancer.install_deepracer_cli(force)
    elif "remove-cli" in sys.argv:
        deepracer.boto3_enhancer.remove_deepracer_cli()


if __name__ == "__main__":
    main()



"""
Copyright 2018-2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
Copyright 2019-2020 AWS DeepRacer Community. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import tarfile
from enum import Enum
from io import BytesIO, TextIOWrapper
from urllib.request import urlopen

import boto3
import pandas as pd

from deepracer import boto3_enhancer
from deepracer.logs import SimulationLogsIO


class LeaderboardSubmissionType(Enum):
    RANKED = 1
    LATEST = 2


class ConsoleHelper:
    """ Class used to retrieve information from AWS DeepRacer Console.
    """
    def __init__(self, profile: str = None, region: str = 'us-east-1'):

        if profile is not None:
            session = boto3.session.Session(profile_name=profile)
        else:
            session = boto3.session.Session()
        self.dr = boto3_enhancer.deepracer_client(session=session, region_name=region)

    def find_model(self, model_name: str) -> str:
        """Finds the Model ARN based on a model display name

        Arguments:
        model_name - (str) String with the model name

        Returns:
        String with the model ARN. None if not found.
        """

        m_response = self.dr.list_models(ModelType="REINFORCEMENT_LEARNING", MaxResults=25)
        model_dict = m_response["Models"]
        models = pd.DataFrame.from_dict(model_dict)
        my_model = models[models["ModelName"] == model_name]

        if my_model.size > 0:
            return my_model.loc[:, 'ModelArn'].values[0]

        while "NextToken" in m_response:
            m_response = self.dr.list_models(
                ModelType="REINFORCEMENT_LEARNING",
                MaxResults=50,
                NextToken=m_response["NextToken"],
            )
            model_dict = m_response["Models"]

            models = pd.DataFrame.from_dict(model_dict)
            my_model = models[models["ModelName"] == model_name]
            if my_model.size > 0:
                return my_model.loc[:, 'ModelArn'].values[0]

        return None

    def find_leaderboard(self, leaderboard_guid: str) -> str:
        """Finds the Leaderboard ARN based on a short GUID

        Arguments:
        leaderboard_guid - (str) String with the leaderboard GUID
                Example: 8fab8c08-66cd-4170-a0b0-ab2b464e42d9

        Returns:
        String with the model ARN. None if not found.
        """

        leaderboard_arn = "arn:aws:deepracer:::leaderboard/{}".format(leaderboard_guid)

        l_response = self.dr.list_leaderboards(MaxResults=25)
        lboards_dict = l_response["Leaderboards"]
        leaderboards = pd.DataFrame.from_dict(l_response["Leaderboards"])
        if leaderboards[leaderboards["Arn"] == leaderboard_arn].size > 0:
            return leaderboard_arn

        while "NextToken" in l_response:
            l_response = self.dr.list_leaderboards(
                MaxResults=50, NextToken=l_response["NextToken"]
            )
            lboards_dict = l_response["Leaderboards"]

            leaderboards = pd.DataFrame.from_dict(lboards_dict)
            if leaderboards[leaderboards["Arn"] == leaderboard_arn].size > 0:
                return leaderboard_arn

        return None

    def get_training_job(self, model_arn: str) -> dict:
        """Retrieves Training Job information for a Model

        Arguments:
        model_arn - (str) String with the Model ARN

        Returns:
        Dict with the Training Job information.
        """

        m_response = self.dr.list_training_jobs(ModelArn=model_arn)
        m_jobs = m_response["TrainingJobs"]
        training_job_arn = m_jobs[0]['JobArn']
        m_response = self.dr.get_training_job(TrainingJobArn=training_job_arn)
        m_job = m_response['TrainingJob']
        return m_job

    def get_training_log_robomaker(self, model_arn: str, data: list = None) -> pd.DataFrame:
        """Retrieves the Robomaker training log for a completed training

        Arguments:
        model_arn - (str) String with the Model ARN
        data - (list) Optional pass in a pre-filled data object

        Returns:
        Pandas DataFrame with the parsed log.
        """
        if data is None:
            data = []

        training_job = self.get_training_job(model_arn)
        training_job_arn = training_job['JobArn']
        f_url = self.dr.get_asset_url(Arn=training_job_arn, AssetType="LOGS")["Url"]
        bytes_io = BytesIO(urlopen(f_url).read())

        with tarfile.open(fileobj=bytes_io, mode="r:gz") as tar_file:
            for member in tar_file.getmembers():
                if member.name.find("robomaker") > 0:
                    log_buf = TextIOWrapper(tar_file.extractfile(member))
                    data = SimulationLogsIO.load_buffer(log_buf)
                    df = SimulationLogsIO.convert_to_pandas(data)
                    return df

        return None

    def get_leaderboard_log_robomaker(self, leaderboard_guid: str,
                                      select=LeaderboardSubmissionType.RANKED,
                                      data: list = None) -> pd.DataFrame:
        """Retrieves the Robomaker log for leaderboard submission

        Arguments:
        leaderboard_guid - (str) String with the Leaderboard GUID
        select - (LeaderboardSubmissionType) Get ranked or latest submission
        data - (list) Optional pass in a pre-filled data object

        Returns:
        Pandas DataFrame with the parsed log.
        """
        if data is None:
            data = []

        leaderboard_arn = self.find_leaderboard(leaderboard_guid)

        if select == LeaderboardSubmissionType.RANKED:
            response_m = self.dr.get_ranked_user_submission(LeaderboardArn=leaderboard_arn)
            submission = response_m["LeaderboardSubmission"]
            activity_arn = "arn:aws:deepracer:us-east-1:180406016328:leaderboard_evaluation_job/{}"\
                .format(submission["SubmissionVideoS3path"].split("/")[7].split("-", 2)[2])

        if leaderboard_arn is None:
            return None

        else:
            response_m = self.dr.get_latest_user_submission(LeaderboardArn=leaderboard_arn)
            submission = response_m["LeaderboardSubmission"]
            activity_arn = submission["ActivityArn"]

        if submission["LeaderboardSubmissionStatusType"] == "SUCCESS":
            f_url = self.dr.get_asset_url(Arn=activity_arn, AssetType="LOGS")["Url"]
            bytes_io = BytesIO(urlopen(f_url).read())

            with tarfile.open(fileobj=bytes_io, mode="r:gz") as tar_file:
                for member in tar_file.getmembers():
                    if member.name.find("robomaker") > 0:
                        log_buf = TextIOWrapper(tar_file.extractfile(member))
                        data = SimulationLogsIO.load_buffer(log_buf)
                        df = SimulationLogsIO.convert_to_pandas(data)
                        df['stream'] = member.name.split("/")[-1].split("-")[0]
                        return df
        else:
            return None



"""
Copyright 2018-2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
Copyright 2020-2021 AWS DeepRacer Community. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import cv2
import numpy as np
import tensorflow as tf
from tensorflow.gfile import GFile


def load_session(pb_path, sensor='FRONT_FACING_CAMERA'):
    sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True,
                                            log_device_placement=True))
    print("load graph:", pb_path)
    with GFile(pb_path, 'rb') as f:
        graph_def = tf.GraphDef()
    graph_def.ParseFromString(f.read())
    sess.graph.as_default()
    tf.import_graph_def(graph_def, name='')
    graph_nodes = [n for n in graph_def.node]
    names = []
    for t in graph_nodes:
        names.append(t.name)

    # For front cameras/stereo camera use the below
    x = sess.graph.get_tensor_by_name(
        'main_level/agent/main/online/network_0/{}/{}:0'.format(sensor, sensor))
    y = sess.graph.get_tensor_by_name('main_level/agent/main/online/network_1/ppo_head_0/policy:0')

    return sess, x, y


def rgb2gray(rgb):
    return np.dot(rgb[..., :3], [0.299, 0.587, 0.114])


def visualize_gradcam_discrete_ppo(
    sess,
    rgb_img,
    sensor='FRONT_FACING_CAMERA',
    category_index=0,
    num_of_actions=5
):
    '''
    @inp: model session, RGB Image - np array, action_index, total number of actions
    @return: overlayed heatmap
    '''

    img_arr = np.array(rgb_img)
    img_arr = rgb2gray(img_arr)
    img_arr = np.expand_dims(img_arr, axis=2)

    x = sess.graph.get_tensor_by_name(
        'main_level/agent/main/online/network_0/{}/{}:0'.format(sensor, sensor))
    y = sess.graph.get_tensor_by_name('main_level/agent/main/online/network_1/ppo_head_0/policy:0')
    feed_dict = {x: [img_arr]}

    # Get he policy head for clipped ppo in coach
    model_out_layer = sess.graph.get_tensor_by_name(
        'main_level/agent/main/online/network_1/ppo_head_0/policy:0')
    loss = tf.multiply(model_out_layer, tf.one_hot([category_index], num_of_actions))
    reduced_loss = tf.reduce_sum(loss[0])

    # For front cameras use the below
    conv_output = sess.graph.get_tensor_by_name(
        'main_level/agent/main/online/network_1/{}/Conv2d_4/Conv2D:0'.format(sensor))

    grads = tf.gradients(reduced_loss, conv_output)[0]
    output, grads_val = sess.run([conv_output, grads], feed_dict=feed_dict)
    weights = np.mean(grads_val, axis=(1, 2))
    cams = np.sum(weights * output, axis=3)

    # im_h, im_w = 120, 160##
    im_h, im_w = rgb_img.shape[:2]

    cam = cams[0]  # img 0
    image = np.uint8(rgb_img[:, :, ::-1] * 255.0)  # RGB -> BGR
    cam = cv2.resize(cam, (im_w, im_h))  # zoom heatmap
    cam = np.maximum(cam, 0)  # relu clip
    heatmap = cam / np.max(cam)  # normalize
    cam = cv2.applyColorMap(np.uint8(255 * heatmap), cv2.COLORMAP_JET)  # grayscale to color
    cam = np.float32(cam) + np.float32(image)  # overlay heatmap
    cam = 255 * cam / (np.max(cam) + 1E-5)  # Add expsilon for stability
    cam = np.uint8(cam)[:, :, ::-1]  # to RGB

    return cam



"""
Copyright 2019-2020 AWS DeepRacer Community. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import json
import math
from io import BytesIO
from typing import TypeVar
from urllib.request import urlopen

import boto3
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.figure import Figure


class TrainingMetrics:
    """ Class used to load in training metrics from S3
    """

    def __init__(
            self,
            bucket: str,
            model_name: str = None,
            pattern: str = "{}/metrics/TrainingMetrics{}.json",
            s3_endpoint_url: str = None,
            region: str = None,
            profile: str = None,
            fname: str = None,
            url: str = None,
            training_round: int = 1,
            display_digits_iteration: int = 3,
            display_digits_episode: int = 4,
            display_digits_round: int = 2
    ):
        """Creates a TrainingMetrics object. Loads the first metrics file into a DataFrame if
            model name is provided.

        Arguments:
        bucket - S3 bucket where the metrics file is stored
        model_name - (str) Name of the model that will be loaded. Default None - returns empty
            object.
        pattern - (str) Filename pattern that will be formatted with the model name to create
            the key.
        training_round - (int) Integer value that will be used to distinguish data.
        display_digits_iteration - (int) Number that will define the padding (e.g for round 1,
            iteration 25 the display_digits_iteration=3 would give unique index as 1-025)
        display_digits_episode - (int) Number that will define the padding (e.g for round 1,
            iteration 25 the display_digits_iteration=4 would give unique index as 1-1225)
        s3_endpoint_url - (str) URL for the S3 endpoint
        region - (str) AWS Region for S3
        profile - (str) Local awscli profile to use when connecting
        fname - (str) Read in a file, rather than from S3

        Returns:
        TrainingMetrics object.
        """
        if profile is not None:
            session = boto3.session.Session(profile_name=profile)
            self.s3 = session.resource("s3", endpoint_url=s3_endpoint_url, region_name=region)
        else:
            self.s3 = boto3.resource("s3", endpoint_url=s3_endpoint_url, region_name=region)
        self.max_iteration_strlen = display_digits_iteration
        self.max_episode_strlen = display_digits_episode
        self.max_round_strlen = display_digits_round
        self.metrics = None
        self.bucket = bucket
        self.pattern = pattern
        if model_name is not None:
            data = self._getS3File(bucket, self.pattern.format(model_name, ''), False)
            df = self._loadRound(
                data, training_round
            )
            self.metrics = df

        if fname is not None:
            data = self._getFile(fname, False)
            df = self._loadRound(
                data, training_round
            )
            self.metrics = df

        if url is not None:
            data = self._getUrlFile(url, False)
            df = self._loadRound(
                data, training_round
            )
            self.metrics = df

    def _getFile(self, fname: str, verbose=False) -> dict:
        if verbose:
            print("Reading in file://%s" % (fname))

        f = open(fname)
        data = json.load(f)
        f.close()

        return data

    def _getS3File(self, bucket: str, key: str, verbose=False) -> dict:

        if verbose:
            print("Downloading s3://%s/%s" % (bucket, key))

        bytes_io = BytesIO()
        self.s3.Object(bucket, key).download_fileobj(bytes_io)
        data = json.loads(bytes_io.getvalue())

        return data

    def _getUrlFile(self, url: str, verbose=False) -> dict:
        if verbose:
            print("Downloading %s" % (url))

        bytes_io = BytesIO(urlopen(url).read())
        data = json.loads(bytes_io.getvalue())

        return data

    def _loadRound(self, data: dict, training_round: int = 1, worker: int = 0,
                   verbose=False) -> pd.DataFrame:

        df = pd.read_json(json.dumps(data["metrics"]), orient="records")
        if worker == 0:
            self.episodes_per_iteration = max(df["trial"])

        df["round"] = training_round
        df["iteration"] = (
            ((df["episode"] - 1) / self.episodes_per_iteration)
            .apply(np.floor)
            .astype(int)
        )
        if self.metrics is not None:
            prev_metrics = self.metrics[self.metrics["round"] < training_round]
            if prev_metrics.shape[0] > 0:
                df["master_iteration"] = (
                    max(prev_metrics["master_iteration"]) + 1 + df["iteration"]
                )
            else:
                df["master_iteration"] = df["iteration"]
        else:
            df["master_iteration"] = df["iteration"]
        self.max_iteration_strlen = max(
            len(str(max(df["iteration"]))), self.max_iteration_strlen
        )
        df["r-i"] = (
            df["round"].astype(str).str.pad(width=self.max_round_strlen, side="left", fillchar="0")
            + "-"
            + df["iteration"]
            .astype(str)
            .str.pad(width=self.max_iteration_strlen, side="left", fillchar="0")
        )

        df["r-e"] = (
            df["round"].astype(str)
            + "-"
            + df["episode"]
            .astype(str)
            .str.pad(width=self.max_episode_strlen, side="left", fillchar="0")
        )
        df["worker"] = worker
        df["reward"] = df["reward_score"]
        df["completion"] = df["completion_percentage"]
        df["complete"] = df["episode_status"].apply(
            lambda x: 1 if x == "Lap complete" else 0
        )
        df["time"] = df["elapsed_time_in_milliseconds"] / 1000
        print(
            ("Successfully loaded training round %i for worker %i: Iterations: %i, " +
             "Training episodes: %i, Evaluation episodes: %i")
            % (
                training_round,
                worker,
                max(df["iteration"]) + 1,
                max(df["episode"]),
                df[df["phase"] == "evaluation"].shape[0],
            )
        )
        return df[
            [
                "r-i",
                "round",
                "iteration",
                "master_iteration",
                "episode",
                "r-e",
                "worker",
                "trial",
                "phase",
                "reward",
                "completion",
                "time",
                "complete",
                "start_time",
            ]
        ]

    def addRound(self, model_name: str, fname: str = None, training_round: int = 2,
                 workers: int = 1):
        """Adds a round of training metrics to the data set

        Arguments:
        model_name - (str) Name of the model that will be loaded.
        fname - (str) Read from file not from S3
        training_round - (int) Integer value that will be used to distinguish data.
        workers - (int) Number of separate workers files to be loaded. (Default: 1)
        """

        if fname is not None:
            data = self._getFile(fname, False)
            df = self._loadRound(
                data, training_round
            )

            if self.metrics is not None:
                self.metrics = self.metrics.append(df, ignore_index=True)
            else:
                self.metrics = df

        else:
            for w in range(0, workers):
                if w > 0:
                    worker_suffix = "_{}".format(w)
                else:
                    worker_suffix = ""

                data = self._getS3File(self.bucket, self.pattern.format(model_name, worker_suffix))
                df = self._loadRound(
                    data, training_round, w
                )

                if self.metrics is not None:
                    self.metrics = self.metrics.append(df, ignore_index=True)
                else:
                    self.metrics = df

    def reloadRound(self, model_name: str, fname: str = None, training_round: int = 2,
                    workers: int = 1):
        """Adds a round of training metrics to the data set

        Arguments:
        model_name - (str) Name of the model that will be loaded.
        training_round - (int) Integer value that will be used to distinguish data.
        workers - (int) Number of separate workers files to be loaded. (Default: 1)
        """

        print("Reloading round {}".format(training_round))
        self.metrics = self.metrics[self.metrics['round'] != training_round]

        if fname is not None:
            data = self._getFile(fname, False)
            df = self._loadRound(
                data, training_round
            )

            if self.metrics is not None:
                self.metrics = self.metrics.append(df, ignore_index=True)
            else:
                self.metrics = df
        else:
            for w in range(0, workers):
                if w > 0:
                    worker_suffix = "_{}".format(w)
                else:
                    worker_suffix = ""

                data = self._getS3File(self.bucket, self.pattern.format(model_name, worker_suffix))
                df = self._loadRound(
                    data, training_round, w
                )

                if self.metrics is not None:
                    self.metrics = self.metrics.append(df, ignore_index=True)
                else:
                    self.metrics = df

    def getEvaluation(self) -> pd.DataFrame:
        """Get the Evaluation part of the data.

        Returns:
        Pandas DataFrame containing all evaluation episodes.
        """
        return self.metrics[self.metrics["phase"] == "evaluation"]

    def getTraining(self) -> pd.DataFrame:
        """Get the Training part of the data.

        Returns:
        Pandas DataFrame containing all training episodes.
        """
        return self.metrics[self.metrics["phase"] == "training"]

    def getSummary(self, rounds: list = None, method: str = "mean",
                   summary_index: list = ["r-i", "iteration"], workers: list = None,
                   completedLapsOnly=False) -> pd.DataFrame:
        """Provides summary per iteration. Data for evaluation and training is separated.

        Arguments:
        rounds - (list) Array of round numbers to be plotted.
        method - (str) Statistical value to be calculated. Examples are 'mean', 'median',
            'min' & 'max'. Default: 'mean'.
        summary_index - (list) List of columns to be used as index of summary.
            Default ['r-i','iteration'].
        completedLapsOnly - (boolean) True will include only completed laps in summary.
        rounds - (list) List of rounds to include in the summary. Defaults to all rounds.
        workers - (list) List of workers to include in the summary. Defaults to all workers.

        Returns:
        Pandas DataFrame containing the summary table.
        """
        input_df = self.metrics

        if workers is not None:
            input_df = input_df[input_df["worker"].isin(workers)]

        if rounds is not None:
            input_df = input_df[input_df["round"].isin(rounds)]

        if completedLapsOnly:
            input_df = input_df[input_df["complete"] == 1]

        columns = summary_index + ["reward", "completion", "time", "complete"]
        training_input = input_df[input_df["phase"] == "training"][columns].copy()
        eval_input = input_df[input_df["phase"] == "evaluation"][columns].copy()

        training_gb = training_input.groupby(summary_index)
        training_agg = getattr(training_gb, method)()
        training_agg.columns = [
            "train_reward",
            "train_completion",
            "train_time",
            "train_completed",
        ]

        training_cnt = training_gb.count()
        training_agg['train_episodes'] = training_cnt['complete']

        eval_gb = eval_input.groupby(summary_index)
        eval_agg = getattr(eval_gb, method)()
        eval_agg.columns = [
            "eval_reward",
            "eval_completion",
            "eval_time",
            "eval_completed",
        ]

        eval_cnt = eval_gb.count()
        eval_agg['eval_episodes'] = eval_cnt['complete']

        return pd.concat([training_agg, eval_agg], axis=1, sort=False)

    ListStr = TypeVar('ListStr', list, str)

    def plotProgress(
            self,
            method: ListStr = "mean",
            rolling_average: int = 5,
            figsize: tuple = (12, 5),
            rounds: list = None,
            workers: list = None,
            series: list = [
                ("eval_completion", "Evaluation", "orange"),
                ("train_completion", "Training", "blue"),
            ],
            title: str = "Completion per Iteration ({})",
            xlabel: str = "Iteration",
            ylabel: str = "Percent complete ({})",
            completedLapsOnly: bool = False,
            grid: bool = False,
    ) -> Figure:
        """Plots training progress. Allows selection of multiple iterations.

        Arguments:
        method - (str / list) Statistical value to be calculated. Examples are 'mean', 'median',
            'min' & 'max'. Default: 'mean'.
        rolling_average - (int) Plotted line will be averaged with last number of x iterations.
            Default: 5.
        figsize - (tuple) Matplotlib figsize definition.
        series - (list) List of series to plot, contains tuples containing column in summary to
            plot, the legend title and color of plot. Default:
            [('eval_completion','Evaluation','orange'),('train_completion','Training','blue')]
        title - (string) The title of the diagram. Optional formatting placeholder '{}' for
            the method.
        xlabel - (string) The x label of the diagram.
        ylabel - (string) The y label of the diagram. Optional formatting placeholder '{}'
            for the method.
        completedLapsOnly - (boolean) Include only completed laps in the statistics.
        grid - (boolean) Adds a grid to the plot.
        rounds - (list) List of rounds to include in the summary. Defaults to all rounds.
        workers - (list) List of workers to include in the summary. Defaults to all workers.

        Returns:
        Matplotlib Figure containing the plot.
        """

        plot_methods = []
        if type(method) is not list:
            plot_methods.append(method)
        else:
            plot_methods = method

        fig, axarr_raw = plt.subplots(1, len(plot_methods), figsize=figsize, sharey=True)

        axarr = []
        if type(axarr_raw) is not np.ndarray:
            axarr.append(axarr_raw)
        else:
            axarr = axarr_raw

        for (m, ax) in zip(plot_methods, axarr):
            summary = self.getSummary(method=m, rounds=rounds, workers=workers,
                                      completedLapsOnly=completedLapsOnly)
            labels = max(math.floor(summary.shape[0] / (15 / len(plot_methods))), 1)
            x = []
            t = []
            for i, a in enumerate(summary.index):
                x.append(a[0])
                if i % labels == 0:
                    t.append(a[0])

            for s in series:
                ax.scatter(x, summary[s[0]], s=2, alpha=0.5, color=s[2])
                ax.plot(
                    x,
                    summary[s[0]].rolling(rolling_average, min_periods=1).mean(),
                    label=s[1],
                    color=s[2],
                )
            ax.set_title(title.format(m))
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel.format(m))
            ax.set_xticks(t)

            ax.legend(loc='upper left')

            self.metrics["iteration"].unique()
            if rounds is not None:
                unique_rounds = rounds[1:]
            else:
                unique_rounds = self.metrics["round"].unique()[1:]

            for r in unique_rounds:
                label = "{}-{}".format(
                    str(r).zfill(self.max_round_strlen),
                    "0".zfill(self.max_iteration_strlen)
                )
                ax.axvline(x=label, dashes=[0.25, 0.75], linewidth=0.5, color="black")

            if grid:
                ax.grid()

        plt.show()
        return fig



"""
Copyright 2018-2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
Copyright 2019-2020 AWS DeepRacer Community. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# os tools to list things
from os import listdir
from os.path import isfile, join

import numpy as np
# Shapely Library
from shapely.geometry import Polygon
from shapely.geometry.polygon import LineString


class TrackIO:
    """Utility to help load npy files containing the track waypoints.

    Track files are considered to contain a numpy array of threes of waypoint coordinates.
    Each three consists of two coordinates in meters when transfered to real world tracks.
    So effectively each element in waypoints list is a numpy array of six numbers:
    [ center_x, center_y, inner_x, inner_y, outer_x, outer_y ]
    The first and last waypoint are usually the same (closing the loop), but
    it isn't a clear requirement in terms of tracks definiton.
    Some of the waypoints may be identical, this is a state which is accepted by the simulator.
    """

    def __init__(self, base_path="./tracks"):
        """Create the TrackIO instance.

        Arguments:
        base_path - base path pointing to a folder containing the npy files.
                    default value: "./tracks"
        """
        self.base_path = base_path

    def get_tracks(self):
        """Yields tracks in the base_path
        """
        for f in sorted(listdir(self.base_path)):
            if isfile(join(self.base_path, f)) and f.endswith('.npy'):
                yield f

    def get_track_waypoints(self, track_name):
        """Load track waypoints as an array of coordinates

        Truth be told, it will load just about any npy file without checks,
        as long as it has the npy extension.

        Arguments:
        track_name - name of the track to load. Both just the name and name.npy
                     will be accepted

        Returns:
        A list of elements where each element is a numpy array of six float values
        representing coordinates of track points in meters
        """
        if track_name.endswith('.npy'):
            track_name = track_name[:-4]
        return np.load("%s/%s.npy" % (self.base_path, track_name))

    def load_track(self, track_name):
        """Load track waypoints as a Track object

        No validation is being made on the input data, results of running on npy files
        other than track info will provide undetermined results.

        Arguments:
        track_name - name of the track to load. Both just the name and name.npy
                     will be accepted

        Returns:
        A Track instance representing the track
        """
        if track_name.endswith('.npy'):
            track_name = track_name[:-4]

        waypoints = self.get_track_waypoints(track_name)

        print("Loaded %s waypoints" % waypoints.shape[0])

        return Track(track_name, waypoints)


class TrackBreakdown:
    """Track info available to enrich the action breakdown graphs.
    Not essential, but can help find interesting sections to pay attention to.
    """

    def __init__(self, vert_lines, track_segments, segment_x, segment_y,
                 segment_xerr, segment_yerr):
        # vert_lines are indices of waypoints put on the track in squares to mark a section
        self.vert_lines = vert_lines
        # track segments determine location of descriptions on the right graph, formed of tuple
        # (location along the x axis, location along the y axis, description)
        self.track_segments = track_segments

        # marking of a bottom-left pixel of a segment on the right graph
        self.segment_x = segment_x
        self.segment_y = segment_y

        # boundaries of red rectangles on the right graph
        # how many pixels wide before and after the bottom-left pixel x coordinate
        self.segment_xerr = segment_xerr
        # how many pixels tall below and above the bottom-left pixel y coordinate
        self.segment_yerr = segment_yerr


reinvent2018 = TrackBreakdown(
    vert_lines=[10, 25, 32, 33, 40, 45, 50, 53, 61, 67],
    track_segments=[(15, 100, 'hairpin'),
                    (32, 100, 'right'),
                    (42, 100, 'left'),
                    (51, 100, 'left'),
                    (63, 100, 'left')],

    segment_x=np.array([15, 32, 42, 51, 63]),
    segment_y=np.array([0, 0, 0, 0, 0]),

    segment_xerr=np.array([[5, 1, 2, 1, 2], [10, 1, 3, 2, 4]]),
    segment_yerr=np.array([[0, 0, 0, 0, 0], [150, 150, 150, 150, 150]]))

london_loop = TrackBreakdown(
    vert_lines=[0, 15, 17, 30, 33, 45, 75, 105, 120, 132, 150, 180, 190, 210],
    track_segments=[(0, 100, 'long sharp left'),
                    (17, 90, 'mild right'),
                    (33, 80, 'tight left'),
                    (75, 100, 'mild chicane'),
                    (120, 100, 'short sharp left'),
                    (150, 90, 'left'),
                    (190, 100, 'right')],

    segment_x=np.array([0, 17, 33, 75, 120, 150, 190]),
    segment_y=np.array([0, 0, 0, 0, 0, 0, 0]),

    segment_xerr=np.array(
        [[0, 0, 0, 0, 0, 0, 0], [15, 13, 12, 30, 12, 30, 20]]),
    segment_yerr=np.array(
        [[0, 0, 0, 0, 0, 0, 0], [150, 150, 150, 150, 150, 150, 150]]))

track_breakdown = {'reinvent2018': reinvent2018, 'london_loop': london_loop}


track_meta = {}
track_meta['Austin'] = 'American Hills Speedway'
track_meta['Singapore'] = 'Asia Pacific Bay Loop'
track_meta['Monaco'] = 'European Seaside Circuit'
track_meta['Aragon'] = 'Stratus Loop'
track_meta['Belille'] = 'Cumulo Turnpike'
track_meta['Albert'] = 'Yun Speedway'
track_meta['July_2020'] = 'Roger Raceway'
track_meta['FS_June2020'] = 'Fumiaki Loop'
track_meta['Spain_track'] = 'Circuit de Barcelona-Catalunya'
track_meta['reInvent2019_track'] = 'The 2019 DeepRacer Championship Cup'
track_meta['reinvent_base'] = 're:Invent 2018'
track_meta['AmericasGeneratedInclStart'] = 'Badaal Track'
track_meta['LGSWide'] = 'SOLA Speedway'
track_meta['Vegas_track'] = 'AWS Summit Raceway'
track_meta['Canada_Training'] = 'Toronto Turnpike Training'
track_meta['Canada_Eval'] = 'Toronto Turnpike Eval'
track_meta['Mexico_track'] = 'Cumulo Carrera Training'
track_meta['Mexico_track_eval'] = 'Cumulo Carrera Eval'
track_meta['China_track'] = 'Shanghai Sudu Training'
track_meta['China_eval_track'] = 'Shanghai Sudu Eval'
track_meta['New_York_Track'] = 'Empire City Training'
track_meta['New_York_Eval_Track'] = 'Empire City Eval'
track_meta['Tokyo_Training_track'] = 'Kumo Torakku Training'
track_meta['Virtual_May19_Train_track'] = 'London Loop Training'
track_meta['Bowtie_track'] = 'Bowtie Track'
track_meta['Oval_track'] = 'Oval Track'
track_meta['reInvent2019_wide'] = 're:Invent 2018 Wide'
track_meta['reInvent2019_wide_mirrored'] = 're:Invent 2018 Wide Mirrored'
track_meta['H_track'] = 'H track'
track_meta['Straight_track'] = 'Straight track'


class Track:
    """Track object represents a track.

    I know, right?

    Fields:
    name - name of the track loaded
    waypoints - input list as received by constructor
    center_line - waypoints along the center of the track with coordinates in meters
    inner_border - waypoints along the inner border of the track with coordinates in meters
    outer_border - waypoints along the outer border of the track with coordinates in meters
    road_poly - a polygon representing the track
    """

    def __init__(self, name, waypoints):
        """Create Track object

        Arguments:
        name - name of the track
        waypoints - values from a npy file for the track
        """
        self.name = name
        self.waypoints = waypoints
        self.center_line = waypoints[:, 0:2]
        self.inner_border = waypoints[:, 2:4]
        self.outer_border = waypoints[:, 4:6]

        l_inner_border = LineString(waypoints[:, 2:4])
        l_outer_border = LineString(waypoints[:, 4:6])
        self.road_poly = Polygon(
            np.vstack((l_outer_border, np.flipud(l_inner_border))))

    def size(self):
        """Track bounding box size tuple
        """
        return (
            np.max(self.outer_border[:, 0]) - np.min(self.outer_border[:, 0]),
            np.max(self.outer_border[:, 1]) - np.min(self.outer_border[:, 1])
        )

    def ratio(self):
        """Track size ratio
        """
        s = self.size()
        return s[1] / s[0]


class GeometryUtils:
    """A set of utilities for use with vectors and points in 2D

    The general idea is to have them work with numpy array representation
    of vectors and points, to simplify working with them.

    Functions work with coordinates provided in numpy arrays as input and
    the results will always be in numpy arrays.

    """

    @staticmethod
    def vector(p1, p2):
        """Convert two points into a vector

        Arguments:
        p1 - first point (either np. array or a list of two floats)
        p2 - second point (either np. array or a list of two floats)

        Returns:
        Vector represented by a numpy array
        """
        return np.subtract(p1, p2)

    @staticmethod
    def get_angle(v1, v2):
        """Calculate an angle between two vectors

        Arguments:
        v1 - first vector (a numpy array)
        v2 - second vector (a numpy array)

        Returns:
        The angle size in degrees
        """

        angle = np.math.atan2(np.linalg.det([v1, v2]), np.dot(v1, v2))
        return np.degrees(angle)

    @staticmethod
    def get_vector_length(v):
        """Calculate the length of a vector, as in magnitude, not amount of coordinates

        Arguments:
        v - a vector (a numpy array)

        Returns:
        Length of the vector (float)
        """

        return np.linalg.norm(v)

    @staticmethod
    def normalize_vector(v):
        """Return a vector scaled to length 1.0

        Arguments:
        v - a vector (a numpy array)

        Returns:
        A normalized vector
        """

        return v / GeometryUtils.get_vector_length(v)

    @staticmethod
    def perpendicular_vector(v):
        """Return a vector perpendicular to one provided

        The output vector is rotated 90 degrees counter-clockwise to input

        Arguments:
        v - a vector (a numpy array)

        Returns:
        A vector perpendicular to input vector
        """

        return np.cross(v, [0, 0, -1])[:2]

    @staticmethod
    def crossing_point_for_two_lines(l1_p1, l1_p2, l2_p1, l2_p2):
        """Returns the point of intersection of the lines

        The lines are passing through l1_p1, l1_p2 and l2_p1, l2_p2

        Result is rounded to three decimal places
        Work by Norbu Tsering https://stackoverflow.com/a/42727584

        Arguments:
        l1_p1 - [x, y] a point on the first line
        l1_p2 - [x, y] another point on the first line
        l2_p1 - [x, y] a point on the second line
        l2_p2 - [x, y] another point on the second line

        Returns:
        Numpy array with coordinates of a point where the two lines cross
        """

        s = np.vstack([l1_p1, l1_p2, l2_p1, l2_p2])        # s for stacked
        h = np.hstack((s, np.ones((4, 1))))  # h for homogeneous
        l1 = np.cross(h[0], h[1])           # get first line
        l2 = np.cross(h[2], h[3])           # get second line
        x, y, z = np.cross(l1, l2)          # point of intersection
        if z == 0:                          # lines are parallel
            return [float('inf'), float('inf')]
        return [np.round(x/z, 3), np.round(y/z, 3)]

    @staticmethod
    def get_a_and_b_for_function(p1, p2):
        """Returns a and b for the function equation y = a*x + b

        Just note it won't work for an equation x = b (when p1_x = p2_x)
        I would normally transpose the coordinates when something like this is needed:
        y' = x
        x' = y
        then a' = 0 and b' = whatever, do the maths and transpose back.

        Arguments:
        p1 - first point on the line
        p2 - second point on the line

        Returns:
        A tuple with a and b for function equation y = a*x + b
        """

        a1 = (p1[1] - p2[1]) / (p1[0] - p2[0])
        b1 = p2[1] - a1 * p2[0]
        return a1, b1

    @staticmethod
    def get_a_point_on_a_line_closest_to_point(l1_p1, l1_p2, p):
        """Finds a point on a straight line which is closest to a given point

        Arguments:
        l1_p1 - first point on the line
        l1_p2 - second point on the line
        p - point to which we want to search the closest point on the line
        """
        vector = GeometryUtils.perpendicular_vector(GeometryUtils.vector(l1_p1, l1_p2))
        p2 = p + vector
        crossing_point = GeometryUtils.crossing_point_for_two_lines(
            l1_p1, l1_p2, p, p2)
        return crossing_point

    @staticmethod
    def is_point_roughly_on_the_line(lp1, lp2, p, tolerated_angle=5):
        """Tells you if the point is roughly on a line.

        In practice it calculates two angles:
                * <- p
               / \
              /   \
             /a1 a2\
           -*-------*--
            ^lp1    ^lp2

        If the absolute value of both the angles is under tolerated_angle,
        The point is close enough to the line to be considered to be on it.

        Just note that when lp1 and lp2 are far from each other, 5 degrees might
        mean a lot, so the point may be not as close as you may think.

        Arguments:
        lp1 - line point 1
        lp2 - line point 2
        p - point to test
        tolerated_angle - how big of an angle can we tolerate? Default is 5 degrees
        """
        a1 = GeometryUtils.get_angle(
            GeometryUtils.vector(lp1, lp2),
            GeometryUtils.vector(lp1, p)
        )

        a2 = GeometryUtils.get_angle(
            GeometryUtils.vector(lp2, p),
            GeometryUtils.vector(lp2, p)
        )
        return a1 < 5 and a2 < 5


class TrackPlotter:
    """Utility to help when trying to plot a track
    """
    @staticmethod
    def plot_track(to_plot, show=True):
        """Plot waypoints for the track

        Arguments:
        waypoints - waypoints to be plotted or the Track object
        show - whether to plot straight away - you may chose to add more to plot
               default value: True
        """
        import matplotlib.pyplot as plt

        if isinstance(to_plot, Track):
            to_plot = to_plot.waypoints

        for point in to_plot:
            plt.scatter(point[0], point[1], c="blue")
            plt.scatter(point[2], point[3], c="black")
            plt.scatter(point[4], point[5], c="cyan")

        if show:
            plt.show()



from enum import Enum


class LogFolderType(Enum):
    CONSOLE_MODEL_WITH_LOGS = 0
    DRFC_MODEL_SINGLE_WORKERS = 1
    DRFC_MODEL_MULTIPLE_WORKERS = 2
    DRFC_MODEL_UPLOAD = 4
    UNKNOWN_FOLDER = 3


class LogType(Enum):
    TRAINING = 0
    EVALUATION = 1
    LEADERBOARD = 2
    NOT_DEFINED = 3
